# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


# x.y.YYYYMMDD or x.y.YYYYMMDD.dev0 -- semver
__version__ = "2.0.20231009"
